;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; schematic/read.scm
;;;
;;; This is a line-based source reader. It provides a procedure,
;;; `port-fold-source-sections`, that folds over each commented block of
;;; code read from an input port yielding comment/source code pairs.
;;;
;;; See this project's README for more information, or
;;; schematic-markdown.scm for a simple usage example.
;;;
;;; Copyright (c) 2013-2016, Evan Hanson <evhan@foldling.org>
;;; See LICENSE for details.
;;;

;;;
;;; SRFI-1-ish.
;;;

(define (any pred lst)
  (and (pair? lst)
       (or (pred (car lst))
           (any pred (cdr lst)))))

(define (drop-while pred lst)
  (cond ((null? lst) lst)
        ((pred (car lst))
         (drop-while pred (cdr lst)))
        (else lst)))

;;;
;;; SRFI-13-ish and other string functions.
;;;

(define (string-null? str)
  (zero? (string-length str)))

;; Infix-joins the list of printable objects `lst` with the optional
;; separator `sep` (which may also be any printable object).
(define (string-join lst . sep)
  (let ((str (open-output-string))
        (sep (and (pair? sep) (car sep))))
    (do ((lst lst (cdr lst)))
        ((null? lst) (get-output-string str))
      (display (car lst) str)
      (when (and sep (pair? (cdr lst)))
        (display sep str)))))

;; Removes characters satisfying `pred` from the front of `str`.
(define (string-drop-while pred str)
  (let ((len (string-length str)))
    (do ((i 0 (+ i 1)))
        ((or (= i len)
             (not (pred (string-ref str i))))
         (substring str i len)))))

;; Determines whether `str1` is a prefix of `str2`.
(define (string-prefix? str1 str2)
  (let ((len1 (string-length str1))
        (len2 (string-length str2)))
    (do ((i 0 (+ i 1)))
        ((or (= i len1)
             (= i len2)
             (not (char=? (string-ref str1 i)
                          (string-ref str2 i))))
         (= i len1)))))

;;;
;;; Source reader & comment formatting.
;;;

;; Creates a procedure that determines whether a given string starts
;; with any of the line comment prefixes in the `prefixes` list.
(define (line-comment-predicate prefixes)
  (lambda (str)
    (any (lambda (p) (string-prefix? p str)) prefixes)))

;; Creates a procedure that returns the documentation part of a line
;; comment, or an empty string if there isn't one.
;;
;; The documentation part of a line comment is the text following a
;; prefix in `prefixes` and a single whitespace character. For example,
;; given the comment prefix list `(";;")`:
;;
;;     "foo"     => ""
;;     ";;"      => ""
;;     ";;bar"   => ""
;;     ";;;foo"  => ""
;;     ";; foo"  => "foo"
;;     ";;  foo" => " foo"
;;
(define (line-comment-content-accessor prefixes)
  (lambda (str)
    (let ((len (string-length str)))
      (or (any (lambda (p)
                 (let ((l (string-length p)))
                   (cond ((= l len)
                          (string))
                         ((and (<= l len) (char-whitespace? (string-ref str l)))
                          (substring str (+ l 1) len))
                         (else #f))))
               prefixes)
          (string)))))

;; As input is read, source lines are accumulated in reverse order, so
;; before they're returned as strings they're reversed, trimmed of empty
;; lines and joined by newlines.
(define (format-source-section lst)
  (string-join
   (drop-while string-null? (reverse (drop-while string-null? lst)))
   #\newline))

;; Fold over each source section read in turn from the given `port`.
(define (port-fold-source-sections kons knil comment-prefixes port)
  (let ((comment?        (line-comment-predicate comment-prefixes))
        (comment-content (line-comment-content-accessor comment-prefixes)))
    (let loop ((docs '())
               (code '())
               (acc knil))
      (let ((line (read-line port)))
        (if (eof-object? line)
            (kons (format-source-section docs)
                  (format-source-section code)
                  acc)
            (let ((str (string-drop-while char-whitespace? line)))
              (cond ((not (comment? str))
                     (loop docs (cons line code) acc))
                    ((null? code)
                     (loop (cons (comment-content str) docs) code acc))
                    (else
                     (loop (list (comment-content str))
                           (list)
                           (kons (format-source-section docs)
                                 (format-source-section code)
                                 acc))))))))))
